import requests
import urllib.parse

from pipeline import api


def __handle_response__(response: requests.Response):
    if response.status_code == 404:
        raise Exception(response)


def post(endpoint, json_data):
    headers = {
        "Authorization": "Bearer %s" % api.PIPELINE_API_TOKEN,
        "Content-type": "application/json",
    }

    url = urllib.parse.urljoin(api.PIPELINE_API_URL, endpoint)
    response = requests.post(url, headers=headers, json=json_data)
    __handle_response__(response)
    return response.json()


def get(endpoint):
    headers = {"Authorization": "Bearer %s" % api.PIPELINE_API_TOKEN}

    url = urllib.parse.urljoin(api.PIPELINE_API_URL, endpoint)

    response = requests.get(url, headers=headers)
    __handle_response__(response)
    return response.json()



from typing import List, Optional

from .base import BaseModel


class FileBase(BaseModel):
    name: str


class FileSchema(FileBase):
    id: str
    path: str

    data: Optional[str]  # Put a limit of 20kB on the data

    class Config:
        orm_mode = True


class FileGet(FileBase):
    id: str
    size: int

    class Config:
        orm_mode = True


class FileCreate(FileBase):
    name: Optional[str]
    file_bytes: Optional[str]



from typing import List, Optional

from .base import BaseModel
from .resource import ResourceCreateSchema, ResourceGetSchema


class SlaveCreateSchema(BaseModel):
    slave_ip: str
    slave_name: str
    resources: List[ResourceCreateSchema]


class SlaveGetSchema(BaseModel):
    id: str
    slave_ip: str
    slave_name: str
    resources: List[ResourceGetSchema]

    class Config:
        orm_mode = True



from pipeline.pipeline_schemas.function import FunctionCreate
from pipeline.api.call import post


def upload_function(function_create_schema: FunctionCreate):

    request_result = post("/v2/functions", function_create_schema.dict())

    return request_result



from typing import List, Optional

from .base import BaseModel


class ResourceCreateSchema(BaseModel):
    foreign_id: Optional[str]
    resource_label: str
    resource_type: str


class ResourceGetSchema(BaseModel):
    id: str
    foreign_id: Optional[str]
    resource_label: str
    resource_type: str

    class Config:
        orm_mode = True



from datetime import datetime
import re
from typing import Optional

from pydantic import BaseModel as PydanticModel, validator


class BaseModel(PydanticModel):
    """Base model for all schemas."""

    class Config:
        # Encode datetime objects to JSON as integer timestamps
        json_encoders = {datetime: datetime.timestamp}


class AvatarHolder(BaseModel):
    """Schema mixin for models which may hold avatars.

    See the associated database model `database.models.Base.AvatarHolder`.
    """

    avatar_colour: Optional[str]
    avatar_image_url: Optional[str]

    @validator("avatar_colour")
    def validate_avatar_colour(cls, value: Optional[str]):
        if value is not None and not re.match(r"#[0-9a-f]{6}", value.lower()):
            raise ValueError("not a valid #rrggbb colour")
        return value



from enum import Enum
import datetime
from typing import List, Optional, Union

from pydantic import Field, validator

from .base import BaseModel
from .data import DataGet
from .runnable import (
    FunctionGet,
    FunctionDetailedGet,
    RunnableIOGet,
    PipelineGet,
    PipelineDetailedGet,
)
from .tag import Tag
from .token import Token


class RunState(Enum):
    RECEIVED = "received"
    ALLOCATING_CLUSTER = "allocating_cluster"
    ALLOCATING_RESOURCE = "allocating_resource"
    LOADING_DATA = "loading_data"
    LOADING_FUNCTION = "loading_function"
    LOADING_PIPELINE = "loading_pipeline"
    RUNNING = "running"
    COMPLETE = "complete"
    FAILED = "failed"


class RunCreateSchema(BaseModel):
    # id: Optional[str]
    # token: Optional[str]
    pipeline_id: Optional[str]
    function_id: Optional[str]
    data: Optional[str]
    data_id: Optional[str]

    class Config:
        orm_mode = True

    @validator("pipeline_id")
    def validate_pipeline(cls, value, **kwargs):
        return value

    @validator("function_id")
    def validate_function(cls, value, **kwargs):
        if (
            "values" in kwargs
            and "pipeline_id" in kwargs["values"]
            and kwargs["values"]["pipeline_id"] != None
        ):
            raise ValueError("Can only pass in function_id or pipeline_id, not both.")
        elif value == None:
            raise ValueError("Must pass in function_id or pipeline_id")
        return value

    @validator("data")
    def validate_data(cls, value, **kwargs):
        return value

    @validator("data_id")
    def validate_data_id(cls, value, **kwargs):
        if (
            "values" in kwargs
            and "data" in kwargs["values"]
            and kwargs["values"]["data"] != None
        ):
            raise ValueError("Can only pass in data or data_id, not both.")
        elif value == None:
            raise ValueError("Must pass in data or data_id")
        return value


class RunIOGet(RunnableIOGet):
    """Realised/given input/output data to a Runnable used in a Run."""

    value: str
    data_url: str


class RunGet(BaseModel):
    id: str
    created_at: datetime.datetime
    started_at: Optional[datetime.datetime]
    ended_at: Optional[datetime.datetime]
    status: RunState = Field(..., alias="run_state")
    compute_time_ms: Optional[int]
    runnable: Union[FunctionGet, PipelineGet]
    # token: str
    data: DataGet
    blocking: Optional[bool] = False

    class Config:
        orm_mode = True
        allow_population_by_field_name = True


class RunDetailedGet(RunGet):
    runnable: Union[FunctionDetailedGet, PipelineDetailedGet]
    n_resources: int
    resource_type: str
    region: str
    tags: List[Tag] = []
    inputs: List[RunIOGet] = []
    outputs: List[RunIOGet] = []
    token: Token


class RunOnResourceSchema(RunGet):
    resource_id: str

    class Config:
        orm_mode = True


"""
class RunCreateSchema(RunSchema):
    token: str
    run_target_type: RunTypeEnum


class RunFunction(BaseModel):
    id: str
    token: str
    function: FunctionGet
"""



import inspect
import requests
import urllib.parse
from dill import dumps

from pipeline.logging import _print
from pipeline.objects import PipelineFunction

from pipeline.api.function import upload_function


PIPELINE_API_TOKEN: str = None
PIPELINE_API_URL: str = "https://api.pipeline.ai"


def authenticate(token: str, url: str = PIPELINE_API_URL):
    """
    Authenticate with the pipeline.ai API
    """
    _print("Authenticating")

    global PIPELINE_API_TOKEN
    PIPELINE_API_TOKEN = token
    global PIPELINE_API_URL
    PIPELINE_API_URL = url
    # TODO: Change this url to an actual auth one, not status which just shows if the API is alive.
    status_url = urllib.parse.urljoin(url, "/v2/status")

    response = requests.get(status_url)
    if response.json()["alive"]:
        _print(
            "Succesfully authenticated with the Pipeline API (%s)" % PIPELINE_API_URL
        )


def upload(object):

    if (
        hasattr(object, "__function__")
        and hasattr(object.__function__, "__pipeline_function__")
        and isinstance(object.__function__.__pipeline_function__, PipelineFunction)
    ):
        function: PipelineFunction = object.__function__.__pipeline_function__

        return upload_function(function._api_create_schema)
    else:
        raise Exception("Not a pipeline object!")



import os
import string
import random
import inspect

from dill import dumps

from hashlib import sha256

from pydantic import BaseModel

from typing import List, Optional, Callable, Any

from pipeline.objects import PipelineFunction


class PipelineVariableSchema(BaseModel):
    id: str

    variable_type: Any
    variable_type_file_path: Optional[str]
    variable_name: str
    is_input: bool = False
    is_output: bool = False
    variable_name: str

    def __init__(self, **kwargs):
        if not "variable_name" in kwargs:
            kwargs["variable_name"] = "".join(
                random.choice(string.ascii_lowercase) for i in range(10)
            )
        kwargs["id"] = "".join(random.choice(string.ascii_lowercase) for i in range(20))
        super().__init__(**kwargs)

    def json(self):
        # print("coming in")
        return self.dict()

    def dict(self, *args, **kwargs):
        # print("Things")
        return {
            "id": self.id,
            "variable_type": str(self.variable_type.__name__),
            "variable_type_file_path": self.variable_type_file_path,
            "variable_name": self.variable_name,
            "is_input": self.is_input,
            "is_output": self.is_output,
        }


class PipelineModel(BaseModel):
    id: str
    remote_id: Optional[str]

    model: Any
    name: str
    hash: Optional[str]
    model_file_name: Optional[str]

    def __init__(self, *args, **kwargs):
        kwargs["id"] = "".join(random.choice(string.ascii_lowercase) for i in range(20))
        if not "name" in kwargs:
            kwargs["name"] = "".join(
                random.choice(string.ascii_lowercase) for i in range(10)
            )
        super().__init__(**kwargs)

    def dict(self, *args, **kwargs):
        return {
            "id": self.id,
            "remote_id": self.remote_id,
            "name": self.name,
            "hash": self.hash,
            "model_file_name": self.model_file_name,
        }


"""
class PipelineFunctionSchema(BaseModel):
    id: str
    remote_id: Optional[str]

    name: str
    hash: Optional[str]
    inputs: dict

    function_hex: str
    function_source: str

    function: Optional[Callable]
    function_file_name: Optional[str]

    bound_class: Optional[Any]
    bound_class_file_name: Optional[str]

    def __init__(self, **kwargs):
        kwargs["id"] = "".join(random.choice(string.ascii_lowercase) for i in range(20))

        function_source = inspect.getsource(kwargs["function"])
        kwargs["hash"] = sha256(function_source.encode()).hexdigest()
        super().__init__(**kwargs)

    def dict(self, *args, **kwargs):
        return {
            "id": self.id,
            "remote_id": self.remote_id,
            "inputs": {
                input_name: self.inputs[input_name].__name__
                for input_name in self.inputs
            },
            "name": self.name,
            "hash": self.hash,
            "function_file_name": self.function_file_name,
            "bound_class_file_name": self.bound_class_file_name,
        }
"""


class PipelineGraphNodeSchema(BaseModel):
    id: str
    pipeline_function: str
    inputs: List[str]
    outputs: List[str]

    def __init__(self, *args, **kwargs):
        kwargs["id"] = "".join(random.choice(string.ascii_lowercase) for i in range(20))
        super().__init__(*args, **kwargs)


class PipelineGraph(BaseModel):
    id: str
    remote_id: Optional[str]

    name: str

    functions: List[PipelineFunction]
    variables: List[PipelineVariableSchema]

    graph_nodes: List[PipelineGraphNodeSchema]

    models: List[PipelineModel]

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, *args, **kwargs):
        kwargs["id"] = "".join(random.choice(string.ascii_lowercase) for i in range(20))
        super().__init__(*args, **kwargs)

    def run(self, *inputs):

        input_variables: List[PipelineVariableSchema] = [
            var for var in self.variables if var.is_input
        ]
        output_variables: List[PipelineVariableSchema] = [
            var for var in self.variables if var.is_output
        ]

        for model in self.models:
            if hasattr(model.model, "load"):
                model.model.load(None)

        if len(inputs) != len(input_variables):
            raise Exception(
                "Mismatch of number of inputs, expecting %u got %s"
                % (len(input_variables), len(inputs))
            )
        running_variables = {}
        for i, input in enumerate(inputs):
            if not isinstance(input, input_variables[i].variable_type):
                raise Exception(
                    "Input type mismatch, expceted %s got %s"
                    % (
                        input_variables[i].variable_type,
                        input.__class__,
                    )
                )
            running_variables[input_variables[i].variable_name] = input

        for node in self.graph_nodes:

            node_inputs: List[PipelineVariableSchema] = []
            node_outputs: List[PipelineVariableSchema] = []
            node_function: PipelineFunction = None

            for _node_input in node.inputs:
                for variable in self.variables:
                    if variable.variable_name == _node_input:
                        node_inputs.append(variable)
                        break
            for _node_output in node.outputs:
                for variable in self.variables:
                    if variable.variable_name == _node_output:
                        node_outputs.append(variable)
                        break

            for function in self.functions:
                if function.name == node.pipeline_function:
                    node_function = function
                    break

            function_inputs = []
            for _input in node_inputs:
                function_inputs.append(running_variables[_input.variable_name])

            if (
                hasattr(node.pipeline_function, "bound_class")
                and node_function.bound_class != None
            ):
                output = node_function.function(
                    node_function.bound_class, *function_inputs
                )
            else:
                output = node_function.function(*function_inputs)

            running_variables[node_outputs[0].variable_name] = output

        return_variables = []
        for output_variable in output_variables:
            return_variables.append(running_variables[output_variable.variable_name])

        return return_variables

    def save(self, save_dir, tar=False, overwrite=True):

        base_save_path = os.path.join(save_dir, "%s" % self.name)
        if os.path.exists(base_save_path):
            if not overwrite and len(os.listdir(base_save_path)) > 0:
                raise Exception("Save directory not empty.")
            elif overwrite and len(os.listdir(base_save_path)) > 0:
                for file in os.listdir(base_save_path):
                    os.remove(os.path.join(base_save_path, file))

        print("Saving pipeline in: '%s'" % base_save_path)
        os.makedirs(base_save_path, exist_ok=True)

        # Save functions
        for node in self.graph_nodes:
            pipeline_function = node.pipeline_function
            function_data = dumps(pipeline_function.function)
            function_source = inspect.getsource(pipeline_function.function)
            function_hash = sha256(function_source.encode()).hexdigest()
            save_file_name = function_hash + ".fn"

            # Update node info
            node.pipeline_function.hash = function_hash
            node.pipeline_function.function_file_name = save_file_name

            with open(
                os.path.join(base_save_path, save_file_name), "wb"
            ) as node_function_file:
                node_function_file.write(function_data)
            if (
                hasattr(node.pipeline_function, "bound_class")
                and node.pipeline_function.bound_class != None
            ):
                bound_class_data = dumps(node.pipeline_function.bound_class)
                # bound_class_source = inspect.getsource(
                #    node.pipeline_function.bound_class
                # )
                bound_class_hash = sha256(bound_class_data).hexdigest()
                bound_class_file_name = bound_class_hash + ".cls"
                node.pipeline_function.bound_class_file_name = bound_class_file_name
                with open(
                    os.path.join(base_save_path, bound_class_file_name), "wb"
                ) as bound_class_file:
                    bound_class_file.write(bound_class_data)
        # Save models
        for model in self.models:
            model_data = dumps(model.model)
            model_hash = sha256(model_data).hexdigest()
            model.hash = model_hash
            model_file_name = model_hash + ".mdl"
            model.model_file_name = model_file_name
            with open(
                os.path.join(base_save_path, model_file_name), "wb"
            ) as model_file:
                model_file.write(model_data)

        # Save variables
        for variable in self.variables:
            type_data = dumps(variable.variable_type)
            type_hash = sha256(type_data).hexdigest()
            type_file_name = type_hash + ".typ"
            variable.variable_type_file_path = type_file_name
            with open(os.path.join(base_save_path, type_file_name), "wb") as type_file:
                type_file.write(type_data)

        with open(os.path.join(base_save_path, "config.json"), "w") as config_file:

            config_file.write(self.json())



import string
import random
import inspect

from dill import dumps

from hashlib import sha256

from typing import Optional, Callable, Any, Union

from pipeline.pipeline_schemas.function import FunctionCreate


class PipelineFunction(object):
    local_id: str = None
    api_id: Optional[str] = None

    name: str = None
    hash: Optional[str] = None
    inputs: dict = {}
    output: dict = {}

    function_hex: str = None
    function_source: str = None

    function: Optional[Callable] = None
    function_file_name: Optional[str] = None

    bound_class: Optional[Any] = None
    bound_class_file_name: Optional[str] = None

    def __init__(self, function: Callable):
        self.local_id = "".join(
            random.choice(string.ascii_lowercase) for i in range(20)
        )
        self.name = function.__name__

        self.inputs = {
            function_i: function.__annotations__[function_i]
            for function_i in function.__annotations__
            if not function_i == "return"
        }

        self.output = {"return": function.__annotations__["return"].__name__}

        self.function = function

        self.function_source = inspect.getsource(function)
        self.function_hex = dumps(function).hex()
        self.hash = sha256(self.function_source.encode()).hexdigest()

    def dict(self, *args, **kwargs):
        return dict(
            local_id=self.local_id,
            remote_id=self.remote_id,
            inputs=self.inputs,
            name=self.name,
            hash=self.hash,
            function_file_name=self.function_file_name,
            bound_class_file_name=self.bound_class_file_name,
        )

    def json(self, *args, **kwargs):
        return self.dict()

    @property
    def _api_create_schema(self) -> FunctionCreate:
        create_schema = FunctionCreate(
            name=self.name,
            hash=self.hash,
            function_hex=self.function_hex,
            function_source=self.function_source,
            inputs={_input: self.inputs[_input].__name__ for _input in self.inputs},
            output=self.output,
        )
        return create_schema

    def _set_bound_class(self, bound_class) -> None:
        self.bound_class = bound_class



import sys
from datetime import datetime

LOG_FILE = None


class bcolors:
    PURPLE = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    ORANGE = "\33[38;5;208m"


levels = {"WARNING": bcolors.ORANGE, "INFO": bcolors.PURPLE, "ERROR": bcolors.FAIL}
PIPELINE_STR = f"[{bcolors.OKBLUE}Pipeline{bcolors.ENDC}]"


def _print(val, level="INFO"):
    time_stamp = datetime.utcnow().strftime("%H:%M:%S")

    log_str = (
        f"{PIPELINE_STR} {time_stamp} - [{levels[level]}{level}{bcolors.ENDC}]: {val}"
    )
    print(f"{log_str}")


def set_print_to_file(path: str):

    global LOG_FILE
    if LOG_FILE == None:
        LOG_FILE = open(path, "w")
        sys.stdout = LOG_FILE
    else:
        raise Exception("Already printing to a log file.")


def stop_print_to_file():

    global LOG_FILE
    if LOG_FILE != None:
        LOG_FILE.close()
    else:
        raise Exception("Not printing to log file")



from datetime import datetime
import enum
from typing import Optional

from .base import BaseModel
from .file import FileSchema
from .token import Token


class FileType(enum.Enum):
    text = "text"
    image = "image"


class DataGet(BaseModel):
    id: str
    hex_file: FileSchema
    created_at: datetime
    modified_at: Optional[datetime]
    name: Optional[str]
    size: Optional[int]
    file_type: Optional[FileType]
    token_created_by: Optional[Token]
    token_modified_by: Optional[Token]
    url: Optional[str]
    preview: Optional[str]

    class Config:
        orm_mode = True



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pipeline',
 'pipeline.api',
 'pipeline.api.function',
 'pipeline.model',
 'pipeline.objects',
 'pipeline.pipeline_schemas',
 'pipeline.pipeline_schemas.redis',
 'pipeline.schemas']

package_data = \
{'': ['*']}

install_requires = \
['black>=21.10b0,<22.0',
 'dill>=0.3.4,<0.4.0',
 'pydantic>=1.8.2,<2.0.0',
 'requests>=2.26.0,<3.0.0',
 'transformers>=4.12.2,<5.0.0']

setup_kwargs = {
    'name': 'pipeline-ai',
    'version': '0.0.8',
    'description': 'Yay ml pipelines',
    'long_description': None,
    'author': 'Paul Hetherington',
    'author_email': 'paul@getneuro.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from .base import BaseModel


class TokenAssignRole(BaseModel):
    role_id: str
    permission_object_id: str


class Token(BaseModel):
    """API token representation when returned from an API call."""

    id: str
    #: Token value, used when authenticating with the API
    value: str
    name: str

    class Config:
        orm_mode = True



from .file import FileSchema
from typing import List, Optional

from .base import BaseModel


class FunctionBase(BaseModel):
    id: Optional[str]
    name: str


class FunctionSchema(FunctionBase):
    id: str


class FunctionGet(FunctionBase):
    id: str
    hex_file: FileSchema
    # source_file_id: FileSchema
    source_sample: str

    class Config:
        orm_mode = True


class FunctionCreate(BaseModel):
    #: The hex-encoded dill-dump of the function object
    function_hex: str
    #: The source code of the function object
    function_source: str
    inputs: dict
    output: dict
    name: str
    hash: str
